import logging
import time
import os
import requests
import json

from werkzeug.exceptions import HTTPException
from werkzeug.http import HTTP_STATUS_CODES
from flask import request, Response, make_response
from decorator import decorator

from ingaia.gcloud import SchedulerUtil, QueueUtil
from .helper import convert_dict_value_to_str

# constants
DB_STORE = True
REQUEST_LOGGER_DATA_STORE_URL = 'https://us-central1-joses-request-logger.cloudfunctions.net/request-logger'

# limit to show the response
RESPONSE_STR_LIMIT = 100


class RequestLogger:
    """
    Request Logger Class
    Save the request and response properties on database
    """

    def __init__(self, module_name=None):
        self._module_name = module_name if module_name else __name__

    def log(self, db_store=DB_STORE, **kwargs):
        """
        Creates a log entry into google cloud datastore

        :param db_store:    defines if store data on database
        :param kwargs:      function parameters
        :return:            request log entry
        """
        func = kwargs.get('func', None)
        func_args = kwargs.get('func_args', None)
        func_resp_status = kwargs.get('func_resp_status', None)
        func_resp = str(kwargs.get('func_resp', None))
        if len(func_resp) > RESPONSE_STR_LIMIT:
            func_resp = func_resp[:RESPONSE_STR_LIMIT] + '[...]'

        headers = None
        request_ip = None
        if request:
            # main headers to store
            headers_to_find = ['Host', 'X-Appengine-User-Ip', 'User-Agent', 'Content-Type', 'X-Forwarded-For',
                               'Forwarded', 'Authorization']
            headers_to_find += SchedulerUtil.CRON_HEADERS
            headers_to_find += QueueUtil.QUEUE_HEADERS
            headers = dict(h for h in request.headers if h[0] in headers_to_find)

            request_ip = request.environ.get('HTTP_X_REAL_IP', request.remote_addr)
            if 'X-Appengine-User-Ip' in headers:
                request_ip = headers['X-Appengine-User-Ip']

        # gets the project id from environment
        project_id = os.environ.get('GCLOUD_PROJECT', 'GCLOUD_PROJECT not set')
        rl_token = os.environ.get('INGAIA_RL_TOKEN', None)

        log_entry = {
            'project_id': project_id,
            'func': func,
            'func_args': func_args,
            'module': self._module_name,
            'request_ip': request_ip,
            'request_url': request.base_url,
            'request_method': request.method,
            'request_headers': headers,
            'request_values': request.values,
            'response': func_resp,
            'response_status': func_resp_status,
            'response_status_name': HTTP_STATUS_CODES.get(func_resp_status, "Unknown Error"),
            'timestamp': int(time.time())
        }

        if db_store and rl_token:
            headers = {
                "Ingaia-Project": project_id,
                "Ingaia-RLToken": rl_token
            }
            requests.post(url=REQUEST_LOGGER_DATA_STORE_URL, json=json.dumps(log_entry), headers=headers)

        return log_entry

    def trace(self, fn, db_store=DB_STORE, *args, **kw):
        """
        @trace - Traces the function responsible for the request (endpoint)

        :param fn:          The function decorated
        :param db_store:    Store the request on database
        :param args:        Args
        :param kw:          KWArgs
        :return:            The same response of the function
        """
        # get function properties
        func_name = f'{fn.__module__}.{fn.__name__}'
        request_url = f'{request.method} {request.base_url}'

        # get the response
        _exception = None
        try:
            _resp = fn(*args, **kw)
        except HTTPException as e:
            _resp = make_response(e, e.code)
            _exception = e

        # verify flask response
        response_status = None
        if isinstance(_resp, Response):
            response_str = _resp.json if _resp.is_json else str(_resp.data)
            response_status = _resp.status_code
        else:
            response_str = str(_resp)
            logging.warning(f'{response_str} (No Flask Response object)')

        log_kwargs = {
            'func': fn.__name__,
            'func_args': {
                'args': convert_dict_value_to_str(list(args)),
                'kwargs': convert_dict_value_to_str(kw)
            },
            'func_resp': response_str,
            'func_resp_status': response_status
        }

        # save the request into google cloud datastore
        log_entry = self.log(db_store, **log_kwargs)

        logging.info(f'[{func_name}] {request_url} :: {response_status} {log_entry}')

        if _exception:
            raise _exception

        return _resp

    def __str__(self):
        return f'{self.__class__.__name__}("{self._module_name}")'


@decorator
def trace(fn, db_store=DB_STORE, *args, **kw):
    return RequestLogger(fn.__module__).trace(fn, db_store=db_store, *args, **kw)



from .exceptions import ProjectReferenceNotDefinedException


class ProjectReference:
    def __init__(self, project_id=None, location=None):
        self._project_id = project_id
        self._location = location

    @property
    def project_id(self):
        if not self._project_id:
            raise ProjectReferenceNotDefinedException()
        return self._project_id

    @project_id.setter
    def project_id(self, project_id):
        self._project_id = project_id

    @property
    def location(self):
        if not self._location:
            raise ProjectReferenceNotDefinedException()
        return self._location

    @location.setter
    def location(self, location):
        self._location = location



"""
Exception classes
"""


class GenericException(Exception):
    message = None

    def __init__(self, message=None):
        self.message = message
        super(GenericException, self).__init__(self.message)


class ProjectReferenceNotDefinedException(GenericException):
    message = 'Define the google cloud project using the "ProjectReference" class.'


class SchemaNotDefinedException(GenericException):
    message = 'Define the bigquery schema.'


class QueryIsNoneException(GenericException):
    message = 'Query is None'


class NoBlobSetException(GenericException):
    message = 'Blob must be set. Use "set_blob([file_name])" method.'


class BigQueryCheckingException(GenericException):
    message = 'There is something wrong with the schema definition or with existing bigquery table.'


class DifferentTableSchemaException(GenericException):
    message = 'The existing table schema is different than the definition schema. ' \
              'To recreate the table, delete that from database.'



import os
import json
import logging

from typing import Dict, List
from google.cloud import bigquery
from .bigquery import BigQueryUtil
from .project import ProjectReference
from google.cloud.bigquery.dataset import DatasetReference
from google.cloud.bigquery.table import TableReference
from google.cloud.bigquery.schema import SchemaField


class SchemaReference:
    def __init__(self, dataset, table, schema):
        self.dataset: DatasetReference = dataset
        self.table: TableReference = table
        self.schema: List[SchemaField] = schema


class SchemaLoader:
    VALID_FIELD_TYPES = ['STRING', 'BYTES', 'INTEGER', 'INT64', 'FLOAT', 'FLOAT64', 'BOOLEAN', 'BOOL',
                         'TIMESTAMP', 'DATE', 'TIME', 'DATETIME', 'RECORD', 'STRUCT']

    def __init__(self, schemas_directory, project=None, project_id=None, location=None):
        self.project = project if project else ProjectReference(project_id, location)
        self._schemas_directory = schemas_directory
        self._json_schemas = None
        self._schemas = None

    @property
    def json_schemas(self) -> list:
        if self._json_schemas is None:
            self._load_directory()
        return self._json_schemas

    def _load_directory(self):
        self._json_schemas = []
        for f in os.listdir(self._schemas_directory):
            file_name, file_extension = os.path.splitext(f)
            if file_extension == '.json' and file_name != '__init__':
                self._load_file(f'{self._schemas_directory}/{f}')

    def _load_file(self, file_path):
        with open(f'{file_path}') as json_file:
            data = json.load(json_file)
            if 'schema' in data:
                schema = data['schema']
                schema['tableReference'] = data['tableReference']
                self._json_schemas.append(schema)

            logging.info(f'The file "{file_path}" was successfully loaded')

    @property
    def schemas(self) -> Dict[str, SchemaReference]:
        if not self._schemas:
            self.load_schemas()
        return self._schemas

    def load_schemas(self) -> Dict[str, SchemaReference]:
        if not self._schemas:
            self._schemas = {}
            for schema in self.json_schemas:
                self._check_schema_and_table(schema)

        return self._schemas

    def load_schema_from_file(self, file_path) -> Dict[str, SchemaReference]:
        self._schemas = {}
        self._json_schemas = []
        self._load_file(file_path)
        self._check_schema_and_table(self._json_schemas[0])
        return self._schemas

    def _check_schema_and_table(self, schema):
        dataset_id = schema['tableReference']['datasetId']
        table_id = schema['tableReference']['tableId']
        reference = f'{dataset_id}.{table_id}'

        bq_schema = self.get_bigquery_schema(schema)

        # check if the definition is the same of the bigquery existing table
        # if the table doesnt exist, it creates one
        bq_util = BigQueryUtil(dataset_id, table_id, schema=bq_schema, project_id=self.project.project_id,
                               location=self.project.location)
        bq_util.check_table()

        # generate the schema reference
        schema_ref = SchemaReference(bq_util.dataset, bq_util.table, bq_util.schema)

        self._schemas[reference] = schema_ref
        logging.info(f"Schema loaded: {schema_ref.table.path}")

    @staticmethod
    def get_bigquery_schema(schema) -> list:
        fields = []
        if 'fields' in schema:
            for field in schema['fields']:
                if 'name' not in field or 'type' not in field:
                    raise Exception('Fields name and type are required')

                field_name = str(field['name']).strip()
                field_type = str(field['type']).strip()  # validate types

                if field_type not in SchemaLoader.VALID_FIELD_TYPES:
                    raise Exception(f'Invalid type: {field_type}')

                field_description = field['description'] if 'description' in field else None
                field_fields = ()
                if field_type == 'RECORD' or field_type == 'STRUCT':
                    if 'fields' not in field:
                        raise Exception('Fields property is required when the type is RECORD or STRUCT')
                    field_fields = tuple(SchemaLoader.get_bigquery_schema({'fields': field['fields']}))

                field_mode = "NULLABLE" if 'mode' not in field else str(field['mode']).strip()
                bq_field = bigquery.SchemaField(field_name, field_type, description=field_description,
                                                mode=field_mode, fields=field_fields)
                fields.append(bq_field)
        return fields
